>From 86f7d6549dc74c1baa58cf9d66f2f2cf4df32666 Mon Sep 17 00:00:00 2001
From: David Fetter <david@fetter.org>
Date: Wed, 25 Feb 2015 13:51:17 -0800
Subject: [PATCH] From Alvaro Herrera

Resolved conflicts:
	src/bin/psql/tab-complete.c
---
 doc/src/sgml/ref/psql-ref.sgml | 31 ++++++++++++++++---
 src/bin/psql/command.c         | 67 +++++++++++++++++++++++++++++-------------
 src/bin/psql/common.c          | 36 +++++++++++++++++++++++
 src/bin/psql/common.h          |  2 ++
 src/bin/psql/tab-complete.c    | 12 +++++++-
 5 files changed, 123 insertions(+), 25 deletions(-)

diff --git a/doc/src/sgml/ref/psql-ref.sgml b/doc/src/sgml/ref/psql-ref.sgml
index 1d034df..d985201 100644
--- a/doc/src/sgml/ref/psql-ref.sgml
+++ b/doc/src/sgml/ref/psql-ref.sgml
@@ -778,18 +778,20 @@ testdb=&gt;
       </varlistentry>
 
       <varlistentry>
-        <term><literal>\c</literal> or <literal>\connect</literal> <literal>[ <replaceable class="parameter">dbname</replaceable> [ <replaceable class="parameter">username</replaceable> ] [ <replaceable class="parameter">host</replaceable> ] [ <replaceable class="parameter">port</replaceable> ] ]</literal></term>
+        <term><literal>\c</literal> or <literal>\connect</literal> <literal> [ { [ <replaceable class="parameter">dbname</replaceable> [ <replaceable class="parameter">username</replaceable> ] [ <replaceable class="parameter">host</replaceable> ] [ <replaceable class="parameter">port</replaceable> ] ] | <replaceable class="parameter">conninfo</replaceable> string | <replaceable class="parameter">URI</replaceable> } ] </literal></term>
         <listitem>
         <para>
         Establishes a new connection to a <productname>PostgreSQL</>
-        server. If the new connection is successfully made, the
-        previous connection is closed. If any of <replaceable
+        server using positional parameters as described below, a
+        <parameter>conninfo</parameter> string, or a <acronym>URI</acronym>. If the new connection is
+        successfully made, the
+        previous connection is closed.  When using positional parameters, if any of <replaceable
         class="parameter">dbname</replaceable>, <replaceable
         class="parameter">username</replaceable>, <replaceable
         class="parameter">host</replaceable> or <replaceable
         class="parameter">port</replaceable> are omitted or specified
         as <literal>-</literal>, the value of that parameter from the
-        previous connection is used. If there is no previous
+        previous connection is used.  If using positional parameters and there is no previous
         connection, the <application>libpq</application> default for
         the parameter's value is used.
         </para>
@@ -804,6 +806,27 @@ testdb=&gt;
         mechanism that scripts are not accidentally acting on the
         wrong database on the other hand.
         </para>
+
+        <para>
+        Positional syntax:
+<programlisting>
+=&gt; \c mydb myuser host.dom 6432
+</programlisting>
+        </para>
+
+        <para>
+        The conninfo form detailed in <xref linkend="LIBPQ-CONNSTRING"> takes two forms: keyword-value pairs
+        </para>
+<programlisting>
+=&gt; \c service=foo
+=&gt; \c "host=localhost port=5432 dbname=mydb connect_timeout=10 sslmode=disable"
+</programlisting>
+        <para>
+        and URIs:
+        </para>
+<programlisting>
+=&gt; \c postgresql://tom@localhost/mydb?application_name=myapp
+</programlisting>
         </listitem>
       </varlistentry>
 
diff --git a/src/bin/psql/command.c b/src/bin/psql/command.c
index 8f8c785..391f60b 100644
--- a/src/bin/psql/command.c
+++ b/src/bin/psql/command.c
@@ -1595,6 +1595,8 @@ do_connect(char *dbname, char *user, char *host, char *port)
 	PGconn	   *o_conn = pset.db,
 			   *n_conn;
 	char	   *password = NULL;
+	bool		keep_password = true;
+	bool		has_connection_string = false;
 
 	if (!o_conn && (!dbname || !user || !host || !port))
 	{
@@ -1608,15 +1610,31 @@ do_connect(char *dbname, char *user, char *host, char *port)
 		return false;
 	}
 
-	if (!dbname)
-		dbname = PQdb(o_conn);
 	if (!user)
 		user = PQuser(o_conn);
+
 	if (!host)
 		host = PQhost(o_conn);
+
 	if (!port)
 		port = PQport(o_conn);
 
+	if (dbname)
+		has_connection_string = recognized_connection_string(dbname);
+
+	keep_password = (
+			(strcmp(user, PQuser(o_conn)) == 0) &&
+			(!host || strcmp(host, PQhost(o_conn)) == 0) &&
+			(strcmp(port, PQport(o_conn)) == 0) &&
+			 !has_connection_string);
+
+	/*
+	 * Unlike the previous stanzas, changing only the dbname shouldn't
+	 * trigger throwing away the password.
+	 */
+	if (!dbname)
+		dbname = PQdb(o_conn);
+
 	/*
 	 * If the user asked to be prompted for a password, ask for one now. If
 	 * not, use the password from the old connection, provided the username
@@ -1631,9 +1649,13 @@ do_connect(char *dbname, char *user, char *host, char *port)
 	{
 		password = prompt_for_password(user);
 	}
-	else if (o_conn && user && strcmp(PQuser(o_conn), user) == 0)
+	else if (o_conn && keep_password)
 	{
-		password = pg_strdup(PQpass(o_conn));
+		password = PQpass(o_conn);
+		if (password && *password)
+			password = pg_strdup(password);
+		else
+			password = NULL;
 	}
 
 	while (true)
@@ -1641,23 +1663,28 @@ do_connect(char *dbname, char *user, char *host, char *port)
 #define PARAMS_ARRAY_SIZE	8
 		const char **keywords = pg_malloc(PARAMS_ARRAY_SIZE * sizeof(*keywords));
 		const char **values = pg_malloc(PARAMS_ARRAY_SIZE * sizeof(*values));
+		int paramnum = 0;
 
-		keywords[0] = "host";
-		values[0] = host;
-		keywords[1] = "port";
-		values[1] = port;
-		keywords[2] = "user";
-		values[2] = user;
-		keywords[3] = "password";
-		values[3] = password;
-		keywords[4] = "dbname";
-		values[4] = dbname;
-		keywords[5] = "fallback_application_name";
-		values[5] = pset.progname;
-		keywords[6] = "client_encoding";
-		values[6] = (pset.notty || getenv("PGCLIENTENCODING")) ? NULL : "auto";
-		keywords[7] = NULL;
-		values[7] = NULL;
+		keywords[0] = "dbname";
+		values[0] = dbname;
+
+		if (!has_connection_string)
+		{
+			keywords[++paramnum] = "host";
+			values[paramnum] = host;
+			keywords[++paramnum] = "port";
+			values[paramnum] = port;
+			keywords[++paramnum] = "user";
+			values[paramnum] = user;
+		}
+		keywords[++paramnum] = "password";
+		values[paramnum] = password;
+		keywords[++paramnum] = "fallback_application_name";
+		values[paramnum] = pset.progname;
+		keywords[++paramnum] = "client_encoding";
+		values[paramnum] = (pset.notty || getenv("PGCLIENTENCODING")) ? NULL : "auto";
+		keywords[++paramnum] = NULL;
+		values[paramnum] = NULL;
 
 		n_conn = PQconnectdbParams(keywords, values, true);
 
diff --git a/src/bin/psql/common.c b/src/bin/psql/common.c
index c08c813..da4f414 100644
--- a/src/bin/psql/common.c
+++ b/src/bin/psql/common.c
@@ -1691,6 +1691,42 @@ session_username(void)
 		return PQuser(pset.db);
 }
 
+/*
+ * Checks if connection string starts with either of the valid URI prefix
+ * designators.
+ *
+ * Returns the URI prefix length, 0 if the string doesn't contain a URI prefix.
+ *
+ * XXX copied from libpq/fe-connect.c
+ */
+static int
+uri_prefix_length(const char *connstr)
+{
+	const char uri_designator[] = "postgresql://";
+	const char short_uri_designator[] = "postgres://";
+	if (strncmp(connstr, uri_designator,
+				sizeof(uri_designator) - 1) == 0)
+		return sizeof(uri_designator) - 1;
+
+	if (strncmp(connstr, short_uri_designator,
+				sizeof(short_uri_designator) - 1) == 0)
+		return sizeof(short_uri_designator) - 1;
+
+	return 0;
+}
+
+/*
+ * Recognized connection string either starts with a valid URI prefix or
+ * contains a "=" in it.
+ *
+ * XXX copied from libpq/fe-connect.c
+ */
+bool
+recognized_connection_string(const char *connstr)
+{
+	return uri_prefix_length(connstr) != 0 || strchr(connstr, '=') != NULL;
+}
+
 
 /* expand_tilde
  *
diff --git a/src/bin/psql/common.h b/src/bin/psql/common.h
index f58c545..069c1b8 100644
--- a/src/bin/psql/common.h
+++ b/src/bin/psql/common.h
@@ -46,4 +46,6 @@ extern const char *session_username(void);
 
 extern void expand_tilde(char **filename);
 
+extern bool recognized_connection_string(const char *connstr);
+
 #endif   /* COMMON_H */
diff --git a/src/bin/psql/tab-complete.c b/src/bin/psql/tab-complete.c
index 9841c1a..3e51ecb 100644
--- a/src/bin/psql/tab-complete.c
+++ b/src/bin/psql/tab-complete.c
@@ -69,6 +69,10 @@ extern char *filename_completion_function();
 /* word break characters */
 #define WORD_BREAKS		"\t\n@$><=;|&{() "
 
+/* XXX Until we can tab complete for URIs and services, each of which
+ * should populate their own lists */
+static const char *const empty_list[] = {NULL};
+
 /*
  * This struct is used to define "schema queries", which are custom-built
  * to obtain possibly-schema-qualified names of database objects.  There is
@@ -3413,7 +3417,13 @@ psql_completion(const char *text, int start, int end)
 /* Backslash commands */
 /* TODO:  \dc \dd \dl */
 	else if (strcmp(prev_wd, "\\connect") == 0 || strcmp(prev_wd, "\\c") == 0)
-		COMPLETE_WITH_QUERY(Query_for_list_of_databases);
+	{
+		/* URI/service completion.  Nothing for now */
+		if (recognized_connection_string(text))
+			COMPLETE_WITH_LIST(empty_list);
+		else
+			COMPLETE_WITH_QUERY(Query_for_list_of_databases);
+	}
 
 	else if (strncmp(prev_wd, "\\da", strlen("\\da")) == 0)
 		COMPLETE_WITH_SCHEMA_QUERY(Query_for_list_of_aggregates, NULL);
-- 
2.1.0

